from skitai.server import counter
from skitai.lib import compressors
from hashlib import md5
from skitai.lib import pathtool
import os
import time
import threading
import glob
import threading
from functools import reduce

class CacheFileSystem:
	def __init__ (self, path, memmax = 64, timeout = 86400):
		self.path = path
		self.memmax = memmax * 1024 * 1024
		self.timeout = timeout
		self.files = {}
		self.hits = {}
		self.memcache = {}
		self.memusage = 0
		self.check_dir ()
		self.lock = threading.RLock ()
		self.numhits = counter.counter ()
		self.numfails = counter.counter ()
	
	def status (self):
		self.lock.acquire ()
		r = {
			"numhits": self.numhits.as_long (),
			"numfails": self.numfails.as_long (),			
			"numchachedfiles": reduce (lambda x, y: x + y, list(self.files.values ()))
		}
		self.lock.release ()
		return r

	def check_dir (self):
		for h in list(range(48,58)) + list(range(97,103)):
			for i in list(range(48,58)) + list(range(97,103)):
				for j in list(range(48,58)) + list(range(97,103)):
					initial = "0" + chr (h) + "/" + chr (i) + chr (j)
					pathtool.mkdir (os.path.join (self.path, initial))
					self.hits [initial] = 1
					self.files [initial] = 0
		
	def maintern (self, initial, force_remove = False):
		c = 0
		for path in glob.glob (os.path.join (self.path, initial, "*")):
			if force_remove or os.stat (path).st_mtime < time.time () - self.timeout:				
				os.remove (path)
			else:
				c += 1
		return c
	
	def truncate (self):
		self.lock.acquire ()
		files = list(self.files.keys ())
		self.lock.release ()
		for initial in files:
			d = self.maintern (initial, True)
			self.lock.acquire ()
			self.files [initial] = c
			self.lock.release ()
						
	def getpath (self, uri, data):
		if not data: 
			data = "" # to make standard type
		key = uri + ":" + str (data)
		file = md5 (key.encode ("utf8")).hexdigest ()
		initial = "0" + file [0] + "/" + file [1:3]
		self.hits [initial] += 1
		if self.hits [initial] % 1000 == 0:
			c = self.maintern (initial)
			self.lock.acquire ()
			self.files [initial] = c
			self.lock.release ()	
		return os.path.join (self.path, initial, file), initial
	
	def decompress (self, content_type, content):
		if content_type [:5] == "text/":
			decompressor = compressors.GZipDecompressor ()
			return decompressor.decompress (content) + decompressor.flush ()
		else:
			content	
			
	def get (self, uri, data, undecompressible = 0):		
		path, initial = self.getpath (uri, data)

		memhit = False
		try:			
			m = self.memcache [initial][path]
			memhit = True
		except KeyError:			
			if not os.path.isfile (path): 
				self.lock.acquire ()
				self.numfails.inc ()
				self.lock.release ()
				return None, None, None, None, None
		
		if memhit:
			mtime, size, max_age = m [0:3]
		else:	
			mtime = os.stat (path).st_mtime
			f = open (path, "rb")
			max_age = int (f.read (12))
		
		if os.stat (path).st_mtime < time.time () - max_age:
			if memhit:
				del self.memcache [initial][path]
				self.memusage -= (size + size * 0.2)
			else:	
				f.close ()
			self.lock.acquire ()
			self.files [initial] -= 1
			self.lock.release ()
			os.remove (path)
			return None, None, None, None, None
		
		self.lock.acquire ()	
		self.numhits.inc ()
		self.lock.release ()
		if memhit:
			compressed, content_type, content = m [3:6]
				
		else:	
			compressed = int (f.read (1))
			content_type = f.read (64).strip ().decode ("utf8")
			content = f.read ()
			f.close ()
			
			if self.memusage < self.memmax:
				size = len (content)
				self.memusage += (size + size * 0.2)
				
				try:
					self.memcache [initial]
				except KeyError:
					self.memcache [initial] = {}
						
				self.memcache [initial][path] = (
					mtime,
					size, 
					max_age,
					compressed, 
					content_type, 
					content
				)
				
		if compressed and not undecompressible:
			decompressor = compressors.GZipDecompressor ()
			content = decompressor.decompress (content) + decompressor.flush ()
			compressed = 0
		
		return memhit and 1 or -1, compressed, max_age, content_type, content
		
	def save (self, uri, data, content_type, content, max_age, compressed = 0):		
		if len (str (max_age)) > 12 or len (content_type) > 64:
			return
		
		path, initial = self.getpath (uri, data)
		if not os.path.isfile (path):
			self.lock.acquire ()
			self.files [initial] += 1
			self.lock.release ()
			f = open (path, "wb")
		else:
			return
		
		if not compressed and content_type.startswith ("text/"):
			compressor = compressors.GZipCompressor ()
			content = compressor.compress (content) + compressor.flush ()
			compressed = 1
		
		if not content_type:
			content_type = b""	
		
		f.write (("%12s%d%64s" % (max_age, compressed, content_type)).encode ("utf8"))
		f.write (content)
		f.close ()
		
		if self.memusage < self.memmax:
			size = len (content)
			self.memusage += (size + size * 0.2)
			
			try:
				self.memcache [initial]
			except KeyError:
				self.memcache [initial] = {}
					
			self.memcache [initial][path] = (
				time.time (),
				size, 
				max_age,
				compressed, 
				content_type, 
				content
			)
			
			
if __name__ == "__main__":
	f = ReverseProxy ("g:\\ttt")
	f.save ("a", "b", "asdasda")
	print(f.get ("a", "b"))
	print(f.status ())



from skitai.saddle import Saddle

def app (env, start_response):
	start_response ("200 OK", [("Content-Type", "text/plain")])
	return ['Hello World!=']

skitaiapp = Saddle (__name__)
skitaiapp.use_reloader = True
skitaiapp.debug = True

@skitaiapp.route('/')
def index (was):
	return "Hello World by Saddle"
	
if __name__ == "__main__":
	import skitai
  
	skitai.run (
		address = "127.0.0.1",
		port = 5000,
		mount = {
			'/': (__file__, 'app'),
			'/skitai': (__file__, 'skitaiapp')
		}
	)




from . import PROXY_TUNNEL_KEEP_ALIVE
import time


class AsynTunnel:
	collector = None
	producer = None
	def __init__ (self, asyncon, handler):
		self.asyncon = asyncon
		self.handler = handler
		self.bytes = 0
		self.asyncon.set_terminator (None)
		
	def collect_incoming_data (self, data):
		#print (">>>>>>>>>>>>", data)
		self.bytes += len (data)		
		self.asyncon.push (data)
	
	def close (self):
		# will be closed by channel
		#print (">>>>>>>>>>>>> AsynTunnel xlosed", self.handler.request.uri)
		self.handler.channel_closed ()
		
	
class TunnelHandler:
	keep_alive = 120
	def __init__ (self, asyncon, request, channel):		
		self.asyncon = asyncon		
		self.request = request
		self.channel = channel
		
		self.asyntunnel = AsynTunnel (asyncon, self)		
		self.channel.set_response_timeout	(PROXY_TUNNEL_KEEP_ALIVE)
		self.asyncon.set_network_delay_timeout (PROXY_TUNNEL_KEEP_ALIVE)
		self.channel.add_closing_partner (self.asyntunnel)
		
		self.bytes = 0
		self.stime = time.time ()
				
	def trace (self, name = None):
		if name is None:
			name = "tunnel://%s:%d" % self.asyncon.address
		self.channel.trace (name)
		
	def log (self, message, type = "info"):
		uri = "tunnel://%s:%d" % self.asyncon.address
		self.channel.log ("%s - %s" % (uri, message), type)
					
	def collect_incoming_data (self, data):	
		#print ("<<<<<<<<<<<<", data[:20])
		self.bytes += len (data)
		self.channel.push (data)
		
	def log_request (self):
		htime = (time.time () - self.stime) * 1000
		self.channel.server.log_request (
			'%s:%d CONNECT tunnel://%s:%d HTTP/1.1 200 %d/%d %dms %dms'
			% (self.channel.addr[0],
			self.channel.addr[1],			
			self.asyncon.address [0],
			self.asyncon.address [1],
			self.asyntunnel is not None and self.asyntunnel.bytes or 0,
			self.bytes,
			htime,
			htime
			)
		)
		
	def connection_closed (self, why, msg):
		#print ("------------Asyncon_disconnected", self.request.uri, self.bytes)
		# Disconnected by server mostly caused by Connection: close header				
		if self.channel:
			self.channel.close_when_done ()
			self.channel.current_request = None
		
	def channel_closed (self):
		#print ("------------channel_closed", self.request.uri, self.bytes)
		self.asyncon.handler = None
		self.asyncon.disconnect ()
		self.asyncon.end_tran ()


from skitai.server import http_response
from skitai.lib import producers
import time


class response (http_response.http_response):
	USE_DATA_COMPRESS = True
	
	def __init__ (self, request):
		http_response.http_response.__init__ (self, request)
	
	def push (self, thing):
		if not self.responsable (): return
		if type(thing) is bytes:			
			self.outgoing.push (producers.simple_producer (thing))
		else:
			self.outgoing.push (thing)
			
	def build_reply_header (self):	
		h = [(b":status", str (self.reply_code).encode ("utf8"))]
		for k, v in self.reply_headers:
			h.append ((k.encode ("utf8"), str (v).encode ("utf8")))		
		return h
	
	def hint_promise (self, uri):
		headers = [
			(':path', uri),
			(':authority', self.request.get_header ('host')),	    
	    (':scheme', self.request.scheme),
	    (':method', "GET")	    
	  ]
	   
		additional_headers = [
	   	(k, v) for k, v in [
		   	('accept-encoding', self.request.get_header ('accept-encoding')),
		   	('accept-language', self.request.get_header ('accept-language')),
		    ('cookie', self.request.get_header ('cookie')),
		    ('user-agent', self.request.get_header ('user-agent')),
		    ('referer', "%s://%s%s" % (self.request.scheme, self.request.get_header ('host'), self.request.uri)),	    
		   ] if v
	   ]
	    
		self.request.http2.push_promise (self.request.stream_id, headers, additional_headers)
		
	def done (self, globbing = True, compress = True, force_close = False, next_request = None):
		# removed by HTTP/2.0 Spec.
		self.delete ('transfer-encoding')
		self.delete ('connection')
		
		if len (self.outgoing) == 0:
			outgoing_producer = None
		
		else:
			way_to_compress = ""
			if self.USE_DATA_COMPRESS and not self.has_key ('Content-Encoding'):
				maybe_compress = self.request.get_header ("Accept-Encoding")
				if maybe_compress and self.has_key ("content-length") and int (self ["Content-Length"]) <= http_response.UNCOMPRESS_MAX:
					maybe_compress = ""
				else:	
					content_type = self ["Content-Type"]
					if maybe_compress and content_type and (content_type.startswith ("text/") or content_type.endswith ("/json-rpc")):
						accept_encoding = [x.strip () for x in maybe_compress.split (",")]
						if "gzip" in accept_encoding:
							way_to_compress = "gzip"
						elif "deflate" in accept_encoding:
							way_to_compress = "deflate"
			
				if way_to_compress:
					if self.has_key ('Content-Length'):
						self.delete ("content-length") # rebuild
					self.update ('Content-Encoding', way_to_compress)
			
			if way_to_compress:
				if way_to_compress == "gzip":
					producer = producers.gzipped_producer
				else: # deflate
					producer = producers.compressed_producer
				outgoing_producer = producer (producers.composite_producer (self.outgoing))						
			else:
				outgoing_producer = producers.composite_producer (self.outgoing)			
			outgoing_producer = producers.globbing_producer (producers.hooked_producer (outgoing_producer, self.log))
		
		try:
			self.request.http2.push_response (
				self.request.stream_id, 
				self.build_reply_header (),
				outgoing_producer
			)
			
		except:
			self.request.logger.trace ()			
			self.request.http2.close (True)		
		else:
			if next_request: # like ssl tunnel
				request, terminator = next_request
				self.request.channel.current_request = request
				self.request.channel.set_terminator (terminator)
	
	def log (self, bytes):		
		self.request.channel.server.log_request (
			'%s:%d %s%s %s %d %dms %dms'
			% (self.request.channel.addr[0],
			self.request.channel.addr[1],			
			self.request.is_promise and "PUSH-" or "",
			self.request.request,
			self.reply_code,			
			bytes,
			self.htime,
			(time.time () - self.stime) * 1000
			)
		)

		
